import os
import shutil
import logging
from typing import List, Optional, Iterator, Set, Type

import jinja2

from .config import Config
from .files import file_class_for_path, ErrorFile
from .jinja import default_extensions, default_filters
from .jinja.exceptions import ReservedVariableError
from .exceptions import BuildError
from .checks.favicon import FaviconCheck
from .checks.issues import Issues
from .files import File


logger = logging.getLogger(__file__)


class Combine:
    def __init__(self, config_path: str, env: str = None, variables: dict = {}) -> None:
        self.config_path = config_path
        self.env = env
        self.variables = variables
        self.load()

    def load(self) -> None:
        self.config = Config(self.config_path)

        self.jinja_environment = self.get_jinja_environment(
            content_paths=self.config.content_paths,
            variables=self.get_jinja_variables(self.config.variables),
        )

        self.content_directories = []
        for path in self.config.content_paths:
            cd = ContentDirectory(path)
            cd.load(self.jinja_environment)
            self.content_directories.append(cd)

    @property
    def output_path(self) -> str:
        return self.config.output_path

    def get_jinja_environment(
        self, content_paths: List[str], variables: dict
    ) -> jinja2.Environment:
        choice_loaders = [jinja2.FileSystemLoader(x) for x in content_paths]

        jinja_environment = jinja2.Environment(
            loader=jinja2.ChoiceLoader(choice_loaders),
            autoescape=jinja2.select_autoescape(["html", "xml"]),
            undefined=jinja2.StrictUndefined,  # make sure variables exist
            extensions=default_extensions,
        )
        jinja_environment.globals.update(variables)
        jinja_environment.filters.update(default_filters)

        return jinja_environment

    def get_jinja_variables(self, config_variables: dict) -> dict:
        """
        1. combine.yml variables
        2. Combine object variables (CLI, Python, etc.) that should override
        3. Built-in variables
        """
        variables = {
            **config_variables,
            **self.variables,
        }

        if "env" in variables:
            raise ReservedVariableError("env")

        variables["env"] = self.env

        return variables

    def reload(self) -> None:
        """Reload the config and entire jinja environment"""
        self.load()

    def clean(self) -> None:
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path)

    def build(self, only_paths: List[str] = [], check: bool = True) -> None:
        build_errors = {}

        if not only_paths:
            # completely wipe it
            self.clean()

        if not os.path.exists(self.output_path):
            os.mkdir(self.output_path)

        paths_rendered = []
        files_rendered = []

        for file in self.iter_files():
            if (
                file.output_relative_path
                and file.output_relative_path not in paths_rendered
            ):
                if only_paths and file.path not in only_paths:
                    continue

                try:
                    file.load(self.jinja_environment)
                    file.render(
                        output_path=self.output_path,
                        jinja_environment=self.jinja_environment,
                    )
                    files_rendered.append(file)
                except Exception as e:
                    build_errors[file.path] = e
                    ErrorFile(file.path, file.content_directory, error=e).render(
                        output_path=self.output_path,
                        jinja_environment=self.jinja_environment,
                    )

                paths_rendered.append(file.output_relative_path)

        if not only_paths:
            self.config.run_build_steps()

        if build_errors:
            for file_path, error in build_errors.items():
                logger.error(f"Error building {file_path}", exc_info=error)
            raise BuildError()

        if check:
            self.check_build(files=files_rendered, site_checks=(not only_paths))

    def check_build(self, files: List[File] = [], site_checks: bool = False) -> None:
        self.issues = Issues()

        if site_checks:
            for issue in FaviconCheck(site_dir=self.output_path).run():
                self.issues.append(issue)

            if self.issues:
                self.issues.print(f"Issues across your site")

        for file in files:
            # TODO could pass check settings here, just don't know what they should look like
            for issue in file.check_output():
                self.issues.append(issue)

    def get_related_files(self, content_relative_path: str) -> List[File]:
        files = []
        for file in self.iter_files():
            if (
                content_relative_path in file.references
                or file.content_relative_path == content_relative_path
            ):
                # TODO could this include duplicates? in the content-relative sense?
                files.append(file)
        return files

    def content_relative_path(self, path: str) -> Optional[str]:
        for content_path in self.config.content_paths:
            if (
                os.path.commonpath([content_path, path]) != os.getcwd()
                and os.getcwd() in content_path
            ):
                return os.path.relpath(path, content_path)

        return None

    def is_in_output_path(self, path: str) -> bool:
        return (
            os.path.commonpath([self.output_path, os.path.abspath(path)]) != os.getcwd()
        )

    def iter_files(self) -> Iterator[File]:
        for content_directory in self.content_directories:
            for file in content_directory.files:
                yield file


class ContentDirectory:
    def __init__(self, path: str) -> None:
        assert os.path.exists(path), f"Path does not exist: {path}"
        self.path = path

    def load(self, jinja_environment: jinja2.Environment) -> None:
        self.files = []

        for root, dirs, files in os.walk(self.path, followlinks=True):
            for file in files:
                file_path = os.path.join(root, file)
                file_obj = file_class_for_path(file_path)(file_path, self)
                file_obj.load(jinja_environment)
                self.files.append(file_obj)

    def file_classes(self) -> Set[Type[File]]:
        return set([x.__class__ for x in self.files])



import jinja2
import os

from .html import HTMLFile
from .utils import create_parent_directory


class RedirectFile(HTMLFile):
    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        redirect_to = open(self.path, "r").read().strip()

        template = jinja_environment.get_template("redirect.template.html")
        with open(target_path, "w+") as f:
            f.write(template.render(redirect_url=redirect_to))

        return target_path



from jinja2 import nodes, BaseLoader
from jinja2.parser import Parser
from jinja2.ext import Extension
from markupsafe import Markup


class IncludeRawExtension(Extension):
    tags = {"include_raw"}

    def parse(self, parser: Parser) -> nodes.Node:
        lineno = parser.stream.expect("name:include_raw").lineno
        template = parser.parse_expression()
        result = self.call_method("_render", [template], lineno=lineno)
        return nodes.Output([result], lineno=lineno)

    def _render(self, filename: str) -> Markup:
        source = self.environment.loader.get_source(self.environment, filename)  # type: ignore
        return Markup(source[0])



import jinja2
import os

import frontmatter

from ..jinja.references import get_references_in_path
from .html import HTMLFile
from .utils import create_parent_directory


class MarkdownFile(HTMLFile):
    def _get_output_relative_path(self) -> str:
        if self.name_without_extension.endswith(".keep"):
            # remove .keep.md from the end and replace with .html
            return super()._get_output_relative_path()[:-8] + ".html"

        if self.name_without_extension == "index":
            return super()._get_output_relative_path()[:-3] + ".html"

        return os.path.join(*self.root_parts, "index.html")

    def load(self, jinja_environment: jinja2.Environment) -> None:
        filename = self._get_jinja_template(
            jinja_environment, self._get_variables()
        ).filename
        if filename:
            self.references = [os.path.basename(filename)] + get_references_in_path(
                filename, jinja_environment
            )
        else:
            self.references = []

    def _get_variables(self) -> dict:
        post = frontmatter.load(self.path)

        variables = post.metadata
        variables["url"] = self._get_url()
        variables["content"] = post.content

        return variables

    def _get_jinja_template(
        self, jinja_environment: jinja2.Environment, variables: dict
    ) -> jinja2.Template:
        template_name = variables.get("template", "markdown.template.html")
        return jinja_environment.get_template(template_name)

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        variables = self._get_variables()

        # TODO can post.content be jinja rendered to use variables?
        template = self._get_jinja_template(jinja_environment, variables)

        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        with open(target_path, "w+") as f:
            f.write(template.render(**variables))

        return target_path



from .core import File
import jinja2


class IgnoredFile(File):
    def _get_output_relative_path(self) -> str:
        return ""

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        return ""



import os

from bs4 import BeautifulSoup
import jinja2
from typing import List

from ..jinja.references import get_references_in_path
from .core import File
from .utils import create_parent_directory
from ..checks.duplicate_id import DuplicateIDCheck
from ..checks.mixed_content import MixedContentCheck
from ..checks.img_alt import ImgAltCheck
from ..checks.meta import MetaDescriptionCheck
from ..checks.title import TitleCheck
from ..checks.links import InternalLinkBrokenCheck
from ..checks.open_graph import (
    OpenGraphTitleCheck,
    OpenGraphDescriptionCheck,
    OpenGraphTypeCheck,
    OpenGraphURLCheck,
    OpenGraphImageCheck,
    OpenGraphSiteNameCheck,
)
from ..checks.base import Check


class HTMLFile(File):
    def _get_output_relative_path(self) -> str:
        if self.name_without_extension.endswith(".keep"):
            # remove .keep.html from the end and replace with .html
            return super()._get_output_relative_path()[:-10] + ".html"

        if self.name_without_extension == "index":
            return super()._get_output_relative_path()

        return os.path.join(*self.root_parts, "index.html")

    def load(self, jinja_environment: jinja2.Environment) -> None:
        self.references = get_references_in_path(self.path, jinja_environment)

    def _render_to_output(
        self, output_path: str, jinja_environment: jinja2.Environment
    ) -> str:
        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        template = jinja_environment.get_template(self.content_relative_path)

        with open(target_path, "w+") as f:
            f.write(template.render(url=self._get_url()))

        return target_path

    def _get_url(self) -> str:
        url = "/" + self.output_relative_path
        if url.endswith("/index.html"):
            url = url[:-10]
        return url

    def get_checks(self) -> List[Check]:
        with open(self.output_path, "r") as f:
            html_soup = BeautifulSoup(f.read(), "html.parser")

            return super().get_checks() + [
                DuplicateIDCheck(html_soup=html_soup),
                MixedContentCheck(html_soup=html_soup),
                ImgAltCheck(html_soup=html_soup),
                MetaDescriptionCheck(html_soup=html_soup),
                TitleCheck(html_soup=html_soup),
                OpenGraphTitleCheck(html_soup=html_soup),
                OpenGraphDescriptionCheck(html_soup=html_soup),
                OpenGraphTypeCheck(html_soup=html_soup),
                OpenGraphURLCheck(html_soup=html_soup),
                OpenGraphImageCheck(html_soup=html_soup),
                OpenGraphSiteNameCheck(html_soup=html_soup),
                InternalLinkBrokenCheck(
                    html_soup=html_soup,
                    file_path=self.output_path,
                    # reverse engineer the output dir for now
                    output_dir=self.output_path[: -len(self.output_relative_path)],
                ),
            ]



import bs4
from .base import Check
from .issues import Issues, Issue


class MetaDescriptionCheck(Check):
    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        meta = self.html_soup.find("meta", {"name": "description"})
        if not meta:
            # missing meta is fine for google, except for social if no og:description (checked elsewhere)
            return issues

        description = meta.get("content", "")
        if not description:
            issues.append(
                Issue(
                    type="meta-description-empty",
                    description="The meta description tag is present, but has an empty value.",
                    context={"element": str(meta)},
                )
            )
            return issues

        if len(description) < 50 or len(description) > 320:
            issues.append(
                Issue(
                    type="meta-description-length",
                    description="The meta description text should be between 50 and 320 characters.",
                    context={"length": len(description), "description": description},
                )
            )

        return issues



from typing import Iterator, List
from pprint import pformat

import click


class Issue:
    def __init__(self, type: str, description: str, context: dict = {}) -> None:
        self.type = type
        self.description = description
        self.context = context

    def as_data(self) -> dict:
        return {
            "type": self.type,
            "context": self.context,
        }


class Issues:
    def __init__(self) -> None:
        self._issues: List[Issue] = []

    def __str__(self) -> str:
        return f"{len(self._issues)} issues"

    def __len__(self) -> int:
        return len(self._issues)

    def __iter__(self) -> Iterator[Issue]:
        return iter(self._issues)

    def append(self, issue: Issue) -> None:
        self._issues.append(issue)

    def print(self, header: str) -> None:
        click.secho("\n" + header, fg="yellow", color=True)
        for index, issue in enumerate(self._issues):
            # TODO could just be yaml format?
            click.secho(
                f"\n  {index+1}. {issue.description}",
                fg="yellow",
                bold=True,
                color=True,
            )
            click.secho(
                f"     https://combine.dropseed.dev/checks/#{issue.type}\n",
                fg="yellow",
                color=True,
            )
            for k, v in issue.context.items():
                click.secho(f"     {k}: {pformat(v)}", fg="yellow", color=True)
        click.echo()

    def as_data(self) -> List[dict]:
        return [issue.as_data() for issue in self._issues]



from typing import List, Iterator, Union
import os
import datetime
import subprocess
import shlex
from fnmatch import fnmatch
import json

import yaml


class Config:
    def __init__(self, path: str) -> None:
        self.path = path
        self.data = {}

        if os.path.exists(self.path):
            self.data = yaml.safe_load(open(self.path, "r"))

    @property
    def output_path(self) -> str:
        return os.path.abspath(self.data.get("output_path", "output"))

    @property
    def content_paths(self) -> List[str]:
        if "content_paths" in self.data:
            paths = self.data["content_paths"]
        else:
            paths = ["content"]
            if os.path.exists(os.path.join("theme", "content")):
                paths.append(os.path.join("theme", "content"))

        # add the built-in content from combine itself
        paths.append(os.path.join(os.path.dirname(__file__), "base_content"))

        return [os.path.abspath(x) for x in paths]

    @property
    def variables(self) -> dict:
        variables = self.default_variables

        for name, data in self.data.get("variables", {}).items():
            # To use a dict as a variable, you'd have to nest it under "default"...
            if isinstance(data, dict):
                if "default" in data:
                    variables[name] = data["default"]

                if "from_env" in data and data["from_env"] in os.environ:
                    variables[name] = os.environ[data["from_env"]]

                if "from_file" in data:
                    filename = data["from_file"]
                    _, ext = os.path.splitext(filename)

                    if not os.path.exists(filename):
                        # Treat it like from_env for now -- if it doesn't exist, that's ok
                        # becuase it will error in use if there isn't a default
                        continue
                    elif ext in (".yml", ".yaml"):
                        variables[name] = yaml.safe_load(open(filename, "r"))
                    elif ext == ".json":
                        variables[name] = json.load(open(filename, "r"))
                    else:
                        # Treat as raw text
                        variables[name] = open(filename, "r").read()
            else:
                variables[name] = data

        return variables

    @property
    def default_variables(self) -> dict:
        return {"now": datetime.datetime.now}  # as a function

    @property
    def steps(self) -> Iterator["BuildStep"]:
        for step in self.data.get("steps", []):
            yield BuildStep(
                run=step.get("run", None),
                watch=step.get("watch", None),
            )

    def run_build_steps(self) -> None:
        for step in self.steps:
            if step.has_run_process:
                step.run_process()


class BuildStep:
    def __init__(self, run: str = None, watch: Union[List[str], str] = None) -> None:
        if watch and not isinstance(watch, (list, str)):
            raise ValueError(
                "watch must be a string (command) or list of strings (paths to watch)"
            )

        self.run = run
        self.watch = watch

    def get_name(self) -> str:
        """
        Get a step name automatically by parsing the first
        executable name in the watch or run command
        """
        if self.has_watch_process:
            return self.watch.split()[0].split("/")[-1]  # type: ignore

        if self.has_run_process:
            return self.run.split()[0].split("/")[-1]  # type: ignore

        return ""

    @property
    def has_watch_process(self) -> bool:
        return isinstance(self.watch, str)

    @property
    def has_watch_patterns(self) -> bool:
        return isinstance(self.watch, list)

    @property
    def has_run_process(self) -> bool:
        return bool(self.run)

    def run_process(self, check: bool = True) -> subprocess.CompletedProcess:
        if not self.run:
            raise ValueError("No run command specified")

        return subprocess.run(shlex.split(self.run), check=check)

    def watch_pattern_match(self, path: str) -> str:
        if not self.has_watch_patterns:
            return ""

        for pattern in self.watch:  # type: ignore
            match = False

            if pattern.startswith("/"):
                # Absolute path pattern
                match = fnmatch(os.path.abspath(path), pattern)
            elif pattern.startswith("./"):
                # Specified relative path pattern
                match = fnmatch(os.path.relpath(path), pattern[2:])
            else:
                # Implied relative path pattern
                match = fnmatch(os.path.relpath(path), pattern)

            if match:
                return pattern

        return ""



import bs4
from .base import Check
from .issues import Issues, Issue


# @Settings.register
# class TitleSuffixSetting(Setting):
#     name = "title-suffix"
#     default = ""


# _titles_seen = set()


class TitleCheck(Check):
    def __init__(self, html_soup: bs4.BeautifulSoup) -> None:
        self.html_soup = html_soup

    def run(self) -> Issues:
        issues = Issues()

        title = self.html_soup.title

        if not title:
            issues.append(
                Issue(type="title-missing", description="The title tag is missing.")
            )
            return issues

        title = title.string

        # if title in _titles_seen:
        #     issues.append(
        #         Issue(
        #             type="duplicate-title", context={"title": title},
        #         )
        #     )
        # else:
        #     _titles_seen.add(title)

        # if len(title.split()) < 2 and not url_is_root(self.url):
        #     issues.append(
        #         Issue(type="title-too-short", context={"title": title},)
        #     )

        # suffix = Settings["title-suffix"]
        # if suffix and title.endswith(suffix):
        #     title = title[: -len(suffix)]

        # if len(title) > 60:
        #     issues.append(Issue(type="title-too-long", context={"title": title}))

        # TODO also check title-empty here? suffix used but template forgot to put something before it...

        if not title:
            issues.append(
                Issue(type="title-empty", description="The title tag has no content.")
            )

        return issues
